# -*- coding: utf-8 -*-

import argparse
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions, SetupOptions, StandardOptions


def run(lambda_class,
        input_subscription,
        output_topic,
        argv=None):
    """Build and run the pipeline."""
    parser = argparse.ArgumentParser()
    known_args, pipeline_args = parser.parse_known_args(argv)

    pipeline_options = PipelineOptions(pipeline_args)
    pipeline_options.view_as(SetupOptions).save_main_session = True
    pipeline_options.view_as(StandardOptions).streaming = True
    p = beam.Pipeline(options=pipeline_options)

    messages = p | beam.io.ReadFromPubSub(subscription=input_subscription)
    messages | beam.ParDo(lambda_class) | beam.io.WriteToPubSub(output_topic)
    result = p.run()
    result.wait_until_finish()



# -*- coding: utf-8 -*-

import json
import uuid
from functools import partial
from datetime import datetime
import apache_beam as beam
from beamism.utils import *


BASE_OUTPUT_SCHEMA = {
    "type": "object",
    "required": ['results', 'is_func_success', 'func_msg'],
    "properties": {
        "is_func_success": { "type": "boolean"},
        "func_msg": { "type": "string"},
        "results": { "type": ["array", "object", "null"] }
    }
}

class Outcome():
    def __init__(self,
                 pipeline='',
                 params={},
                 element_id='',
                 sender_element_id='',
                 data=None,
                 succeeded=False,
                 message='',
                 start_utc_time=None,
                 end_utc_time=None,
                 processing_time=None,
                 func_version=None):
        self.pipeline_name = pipeline
        self.params = params
        self.element_id = element_id
        self.sender_element_id = sender_element_id
        self.data = data
        self.succeeded = succeeded
        self.message = message
        self.start_utc_time = start_utc_time
        self.end_utc_time = end_utc_time
        self.processing_time = processing_time
        self.func_version = func_version

    def to_json(self):
        return {
            "pipeline": self.pipeline_name,
            "params": self.params,
            "element_id": self.element_id,
            "sender_element_id": self.sender_element_id,
            "data": self.data,
            "succeeded": self.succeeded,
            "message": self.message,
            "start_utc_time": self.start_utc_time,
            "end_utc_time": self.end_utc_time,
            "processing_time": self.processing_time,
            "func_version": self.func_version
        }

class Lambda(beam.DoFn):
    '''
    Parameters:
        func:
            function(dict => (dict, bool, str))
        save_func:
            function(dict => bool)
        pipeline_name: str
        func_version: str
    '''
    def __init__(self,
                 func,
                 input_schema,
                 save_func=None,
                 pipeline_name='',
                 func_version=None,
                 notificator=SilentNotificator()):
        self.func = func
        self.input_schema = input_schema
        self.pipeline_name = pipeline_name
        self.save_func = save_func
        self.func_version = func_version
        self.notificator = notificator

    def process(self, msg_text, save_mode=True):
        if type(msg_text) == bytes:
            msg_text = msg_text.decode('utf-8')
        element_id = str(uuid.uuid4()) # Pipelineが処理する一つ一つのJob固有のID( ≠ message_id)
        start_utc_time = datetime.utcnow()
        input_json = parse_json_message(msg_text)
        if input_json is None: # インプットがJSON形式ではないケース
            output_json = self.create_output_json_when_input_is_not_json(start_utc_time, msg_text, element_id)
            self.notificator.parse_error(msg_text)
            self.save(output_json, save_mode)
            return data_packing(output_json)

        self.notificator.start(msg_text, input_json)

        if self.input_schema is not None: # インプットが与えられたスキーマに従っていないケース
            input_validation_err = validate_json_based_on_schema(input_json, self.input_schema)
            if input_validation_err is not None:
                error_msg = "{0}@{1}".format(input_validation_err.message, ''.join(str(input_validation_err.absolute_path)))
                output_json = self.create_output_json_when_validation_fail(start_utc_time, msg_text, element_id, error_msg, input_json)
                self.notificator.validation_error(msg_text, input_json)
                self.save(output_json, save_mode)
                return data_packing(output_json)

        if not "succeeded" in input_json or not input_json["succeeded"]: # 一つ前のJOBが成功していないケース
            output_json = self.create_output_json_when_previous_job_failed(start_utc_time, msg_text, element_id, input_json)
            self.notificator.previous_job_failed(msg_text, input_json)
            self.save(output_json, save_mode)
            return data_packing(output_json)

        if not "data" in input_json or input_json["data"] is None: # データが存在しないケース
            output_json = self.create_output_json_when_data_not_found(start_utc_time, msg_text, element_id, input_json)
            self.notificator.data_not_found(msg_text, input_json)
            self.save(output_json, save_mode)
            return data_packing(output_json)

        indata = input_json["data"]
        main_output = self.func(indata)

        if validate_json_based_on_schema(main_output, BASE_OUTPUT_SCHEMA):
            func_msg = 'Output of main_func should follow the BASE_OUTPUT_SCHEMA. The output is follows: ' + str(main_output)
            results = [{}]
            is_func_success = False
        else:
            results, is_func_success, func_msg = (main_output['results'], main_output['is_func_success'], main_output['func_msg'])

        if type(results) != list:
            results = [{}] if results is None else [results]

        output_json_for_save_func = self.create_output_json_when_msg_is_valid(start_utc_time,
                                                                              msg_text,
                                                                              element_id,
                                                                              input_json,
                                                                              is_func_success,
                                                                              func_msg,
                                                                              results)
        self.save(output_json_for_save_func, save_mode)
        outdatas = [merge_jsons(indata, result) for result in results]
        output_jsons = [self.create_output_json_when_msg_is_valid(start_utc_time,
                                                                  msg_text,
                                                                  element_id,
                                                                  input_json,
                                                                  is_func_success,
                                                                  func_msg,
                                                                  outdata) for outdata in outdatas]
        self.notificator.finish(msg_text, input_json)
        return data_packing(output_jsons)

    def save(self, output_json, save_mode=True):
        if self.save_func is not None and save_mode:
            is_acknowledged = self.save_func(output_json)
            return is_acknowledged
        return False

    def create_output_json_when_input_is_not_json(self, start_utc_time, msg_text, element_id):
        return Outcome(pipeline=self.pipeline_name,
                       params={ "input_text": msg_text, "error": "parse error" },
                       element_id=element_id,
                       data=None,
                       succeeded=False,
                       message="Cannot parse this message as json",
                       start_utc_time=str(start_utc_time),
                       end_utc_time=str(datetime.utcnow()),
                       processing_time=(datetime.utcnow() - start_utc_time).total_seconds(),
                       func_version=self.func_version).to_json()

    def create_output_json_when_validation_fail(self, start_utc_time, msg_text, element_id, error_msg, input_json):
        return Outcome(pipeline=self.pipeline_name,
                       params=input_json["data"],
                       element_id=element_id,
                       data=None,
                       succeeded=False,
                       message=error_msg,
                       start_utc_time=str(start_utc_time),
                       end_utc_time=str(datetime.utcnow()),
                       processing_time=(datetime.utcnow() - start_utc_time).total_seconds(),
                       func_version=self.func_version).to_json()

    def create_output_json_when_previous_job_failed(self, start_utc_time, msg_text, element_id, input_json):
        return Outcome(pipeline=self.pipeline_name,
                       params=input_json,
                       element_id=element_id,
                       sender_element_id=input_json["element_id"] if "element_id" in input_json else None,
                       data=None,
                       succeeded=False,
                       message="Previous job failed",
                       start_utc_time=str(start_utc_time),
                       end_utc_time=str(datetime.utcnow()),
                       processing_time=(datetime.utcnow() - start_utc_time).total_seconds(),
                       func_version=self.func_version).to_json()

    def create_output_json_when_data_not_found(self, start_utc_time, msg_text, element_id, input_json):
        return Outcome(pipeline=self.pipeline_name,
                       params=input_json["data"],
                       element_id=element_id,
                       sender_element_id=input_json["element_id"] if "element_id" in input_json else None,
                       data=None,
                       succeeded=False,
                       message="Data not found",
                       start_utc_time=str(start_utc_time),
                       end_utc_time=str(datetime.utcnow()),
                       processing_time=(datetime.utcnow() - start_utc_time).total_seconds(),
                       func_version=self.func_version).to_json()

    def create_output_json_when_msg_is_valid(self,
                                             start_utc_time,
                                             msg_text,
                                             element_id,
                                             input_json,
                                             is_func_success,
                                             func_msg,
                                             data):
        return Outcome(pipeline=self.pipeline_name,
                       params=input_json["data"],
                       element_id=element_id,
                       sender_element_id=input_json["element_id"] if "element_id" in input_json else None,
                       data=data,
                       succeeded=is_func_success,
                       message=func_msg,
                       start_utc_time=str(start_utc_time),
                       end_utc_time=str(datetime.utcnow()),
                       processing_time=(datetime.utcnow() - start_utc_time).total_seconds(),
                       func_version=self.func_version).to_json()



# -*- coding: utf-8 -*-


import json
import jsonschema
import copy
from pymongo import MongoClient
import pkg_resources
from abc import ABCMeta, abstractmethod


class Notificator():
    __metaclass__ = ABCMeta

    @abstractmethod
    def parse_error(self, msg_text):
        pass

    @abstractmethod
    def start(self, msg_text, input_json):
        pass

    @abstractmethod
    def validation_error(self, msg_text, input_json):
        pass

    @abstractmethod
    def previous_job_failed(self, msg_text, input_json):
        pass

    @abstractmethod
    def data_not_found(self, msg_text, input_json):
        pass

    @abstractmethod
    def finish(self, msg_text, input_json):
        pass

class SilentNotificator(Notificator):
    def __init__(self):
        pass

    def parse_error(self, msg_text):
        pass

    def start(self, msg_text, input_json):
        pass

    def validation_error(self, msg_text, input_json):
        pass

    def previous_job_failed(self, msg_text, input_json):
        pass

    def data_not_found(self, msg_text, input_json):
        pass

    def finish(self, msg_text, input_json):
        pass

def parse_json_message(json_as_str):
    try:
        input_json = json.loads(json_as_str)
        return input_json
    except:
        return None


def validate_json_based_on_schema(input_json, schema):
    try:
        jsonschema.validate(input_json, schema)
        return None
    except jsonschema.ValidationError as validation_error:
        return validation_error


def data_packing(json_datas):
    '''
    How to Pack:
        convert msgs as list(json) => msgs as list[str(json style)]
    Note:
        When return string as message, pubsub treat it as list of chars. But, in many cases, we do not want such behaviour.
    '''
    if type(json_datas) != list:
        json_datas = [json_datas]
    return [json.dumps(json_data).encode() for json_data in json_datas]


def merge_jsons(params_json, result_json):
    """
    Merge two jsons: the first one should be the parameters for the job, and the other should be the results of the job
    """
    # make sure keys not change
    if set(params_json.keys()).intersection(result_json.keys()):
        err_msg = "[BUG] When merge two jsons, some keys are modified"
        print(err_msg)

    merged_json = copy.copy(result_json)
    merged_json.update(params_json)
    return merged_json


def get_version():
    return pkg_resources.get_distribution("beamism").version



# -*- coding: utf-8 -*-
#!/usr/bin/env python

from setuptools import setup, find_packages


REQUIRED_PACKAGES = [
    'apache-beam==2.13.0',
    'jsonschema==2.6.0',
    'dnspython==1.15.0',
    'gapic-google-cloud-pubsub-v1==0.15.4',
    'gitpython==2.0.8',
    'google-api-core==1.1.2',
    'google-apitools==0.5.20',
    'google-auth==1.4.1',
    'google-auth-httplib2==0.0.3',
    'google-cloud-core==0.25.0',
    'google-cloud-pubsub==0.26.0',
    'google-gax==0.15.16',
    'googleapis-common-protos==1.5.3',
    'oauth2client==3.0.0',
    'proto-google-cloud-pubsub-v1==0.15.4',
    'python-dateutil==2.7.3',
    'pymongo==3.6.1',
    'slackclient==1.2.1'
]


with open('version.cache', 'r') as f:
    version = f.read()


setup(
    name="beamism",
    version=version,
    packages=find_packages(),
    description='welcome to beamism',
    platforms='Linux, Darwin',
    zip_safe=False,
    include_package_data=True,
    install_requires=REQUIRED_PACKAGES,
    license="MIT",
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ]
)
